/*
 A variable-algorithm cipher: the operations depend on the key.

 Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2013-01-31 (Public Domain)
*/
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <stdlib.h>

typedef unsigned char cipher_t; // datatype used by cipher
typedef unsigned long cikey_t; // key type used by cipher
typedef cipher_t (*cipher_fn)(cipher_t a,cipher_t b); // one cryptographic operation

/* Destructively extract the low n bits from this key */
cikey_t key_bits(unsigned int nbits,cikey_t &key) {
	cikey_t ret=key&((1<<nbits)-1);
	key=key>>nbits;
	return ret;
}

// These are the basic operations in each stage of the cipher.
cipher_t nop_fn(cipher_t a,cipher_t k) { return a; }
cipher_t add_fn(cipher_t a,cipher_t k) { return a+k; }
cipher_t xor_fn(cipher_t a,cipher_t k) { return a^k; }
cipher_t rot_fn(cipher_t a,cipher_t k) { return (a<<(k&7))|(a>>(8-(k&7))); }

// This lists the possible functions to apply at each stage of the cipher
cipher_fn fn_table[]={ nop_fn, add_fn, xor_fn, rot_fn };
// This is the number of bits of key consumed at this stage (MUST match fn_table!)
unsigned int keybits_table[]={ 0, 8, 8, 3 };

// Apply one stage of this cipher, consuming bits of the key
cipher_t apply_stage(cipher_t data,cikey_t &key) {
	cikey_t stage=key_bits(2,key); // operation to apply
	cipher_fn fn=fn_table[stage]; // function to apply it
	cipher_t k=key_bits(keybits_table[stage],key); // local key
	return fn(data,k);
}

// Keep applying stages of this cipher until there are no more stages
cipher_t apply_stages(cipher_t data,cikey_t key) {
	cikey_t lkey=key; // copy, to be consumed encrypting this byte
	while (lkey!=0) data=apply_stage(data,lkey);
	return data;
}

// Read plaintext, write ciphertext by encrypting each byte with the key (ECB mode).
int main(int argc,char *argv[]) {
	cikey_t key=0;
	if (argc>1) sscanf(argv[1],"%lx",&key); // take key in hex as command-line argument

	const char *fileName="plain.bin";
	std::ifstream file(fileName,std::ios_base::binary);
	std::ofstream of("cipher.bin",std::ios_base::binary);
	unsigned char p;
	while (file.read((char *)&p,1)) {
		unsigned char c=apply_stages(p,key);
		of.write((char *)&c,1);
	}
}

